# -*- coding: utf-8 -*-
import json
from typing import MutableMapping


class JsonSerializer:
    @staticmethod
    def dump(filename: str, tables: MutableMapping):
        with open(filename, "w", encoding="utf-8") as file:
            json.dump(tables, file, ensure_ascii=False)

    @staticmethod
    def load(filename: str) -> MutableMapping:
        with open(filename, "r", encoding="utf-8") as file:
            return json.load(file)



# -*- coding: utf-8 -*-
from enum import auto
from enum import Enum


class ComparisonType(Enum):
    EQUAL = auto()
    NOT_EQUAL = auto()


class Query:
    def __init__(self, field, comparison_type: ComparisonType, value):
        self.field = field
        self.comparison_type = comparison_type
        self.value = value



# -*- coding: utf-8 -*-
from typing import Iterable
from typing import Iterator
from typing import Mapping
from typing import MutableMapping
from uuid import uuid4

from .document import Document
from .field import Field
from .fields import Fields
from .query import ComparisonType
from .query import Query


class Table(Iterable):
    def __init__(self, _fields: Iterable[Field] = None):
        self._data = {}
        self.fields = Fields(_fields=_fields)

    @classmethod
    def _deserialize(cls, _data: MutableMapping):
        table = cls()
        table.fields = Fields._deserialize(_data["_fields"])
        table._data = {
            key: Document(
                _table=table,
                _document_key=value["_document_key"],
                mapping=value["_data"],
            )
            for key, value in _data["_data"].items()
        }
        return table

    def _serialize(self) -> dict:
        return {
            "_data": {
                key: {"_document_key": document._document_key, "_data": dict(document)}
                for key, document in self._data.items()
            },
            "_fields": self.fields._serialize(),
        }

    def __iter__(self):
        for value in self._data.values():
            yield value

    def __len__(self):
        return self._data.__len__()

    def __repr__(self):
        return "Table(%r)" % list(self._data.values())

    def insert(self, document):
        if not isinstance(document, Mapping):
            raise TypeError("Only dict-compatible types are supported")

        key = uuid4().hex
        document = Document(_table=self, _document_key=key, mapping=document)
        self._data.__setitem__(key, document)

    def upsert(self, document, query, multiple=False):
        for key in self._get_keys(query):
            self._data[key].update(document)
            if multiple is False:
                break

    def _get_keys(self, query: Query) -> Iterator:
        if query.field.indexed:
            if query.comparison_type == ComparisonType.EQUAL:
                for key in query.field._get_keys(ComparisonType.EQUAL, query.value):
                    yield key
            elif query.comparison_type == ComparisonType.NOT_EQUAL:
                equal_keys = query.field._get_keys(ComparisonType.EQUAL, query.value)
                for key in self._data.keys():
                    if key not in equal_keys:
                        yield key
        else:
            if query.comparison_type == ComparisonType.EQUAL:
                for key, document in self._data.items():
                    if document.get(query.field.name) == query.value:
                        yield key
            elif query.comparison_type == ComparisonType.NOT_EQUAL:
                for key, document in self._data.items():
                    if document.get(query.field.name) != query.value:
                        yield key

    def get(self, query: Query) -> Iterator:
        for key in self._get_keys(query):
            yield self._data[key]

    def first(self, query: Query):
        for key in self._get_keys(query):
            return self._data[key]
        return None

    def _delete_document(self, document_key):
        self._data[document_key].clear()
        del self._data[document_key]

    def delete(self, query: Query):
        keys = list(self._get_keys(query))
        for key in keys:
            self._delete_document(key)



# -*- coding: utf-8 -*-
from typing import Iterable
from typing import Iterator
from typing import MutableMapping

from .field import Field
from .query import Query
from .serializers import JsonSerializer
from .table import Table


class SmolStore(Iterable):
    def __init__(self, filename=None, serializer=JsonSerializer):
        self._filename = filename
        self._serializer = serializer
        if filename:
            self._load_from_file()
        else:
            self._tables = {}

    def _load_from_file(self):
        try:
            self._deserialize(self._serializer.load(self._filename))
        except FileNotFoundError:
            self._tables = {}

    def _deserialize(self, _data: MutableMapping):
        self._tables = {
            table_name: Table._deserialize(table) for table_name, table in _data.items()
        }

    def _serialize(self) -> dict:
        return {
            table_name: table._serialize() for table_name, table in self._tables.items()
        }

    def table(self, table_name="_default", fields: Iterable[Field] = None):
        if table_name in self._tables:
            return self._tables[table_name]
        else:
            table = Table(_fields=fields)
            self._tables[table_name] = table
            return table

    def __iter__(self):
        return self.table().__iter__()

    def __len__(self):
        return self.table().__len__()

    @property
    def fields(self):
        return self.table().fields

    def save(self):
        if self._filename:
            self._serializer.dump(self._filename, self._serialize())

    def delete_table(self, table_name: str):
        try:
            del self._tables[table_name]
        except KeyError:
            pass

    def insert(self, document):
        self.table().insert(document)

    def upsert(self, document, query, multiple=False):
        self.table().upsert(document, query, multiple)

    def get(self, query: Query) -> Iterator:
        return self.table().get(query)

    def first(self, query: Query):
        return self.table().first(query)

    def delete(self, query: Query):
        self.table().delete(query)



# -*- coding: utf-8 -*-
from typing import Mapping
from typing import MutableMapping

from .exceptions import ReservedKey


class Document(MutableMapping):
    def __init__(self, _table, _document_key, mapping):
        self.__dict__["_reserved_keys"] = {
            "_reserved_keys",
            "_document_table",
            "_document_key",
        }
        self.__dict__["_document_table"] = _table
        self.__dict__["_document_key"] = _document_key
        self.update(mapping)

    def _set_item(self, key, value):
        if key in self["_reserved_keys"]:
            raise ReservedKey("'%s' is a reserved document key in smolstore" % key)

        if not isinstance(value, Mapping):
            old_value = self.get(key)
            self["_document_table"].fields._register(
                document_key=self["_document_key"],
                field_name=key,
                old_value=old_value,
                new_value=value,
            )

        self.__dict__[key] = value

    def _del_item(self, key):
        print("deleted")
        if key in self["_reserved_keys"]:
            raise ReservedKey("'%s' is a reserved document key in smolstore" % key)

        self["_document_table"].fields._unregister(
            document_key=self["_document_key"], field_name=key, value=self.get(key)
        )

        del self.__dict__[key]

    def __setattr__(self, key, value):
        self._set_item(key, value)

    def __delattr__(self, item):
        self._del_item(item)

    def __setitem__(self, key, value):
        self._set_item(key, value)

    def __getitem__(self, key):
        return self.__dict__.__getitem__(key)

    def __delitem__(self, key):
        self._del_item(key)

    def __iter__(self):
        for item in self.__dict__:
            if item not in self["_reserved_keys"]:
                yield item

    def __len__(self):
        return len(dict(self))

    def __repr__(self):
        return "Document(%r)" % dict(self)

    def __str__(self):
        return dict(self).__str__()

    def delete(self):
        self["_document_table"]._delete_document(self["_document_key"])



# -*- coding: utf-8 -*-
from collections import defaultdict

from .exceptions import UniqueViolation
from .query import ComparisonType
from .query import Query


class Field:
    def __init__(self, field_name=None, index=False, unique=False):
        if unique:
            index = True
        self.name = field_name
        self.indexed = index
        self.unique = unique

        self._hash_index = defaultdict(lambda: set())

    @classmethod
    def _deserialize(cls, _data):
        field = cls(**_data)
        return field

    def _serialize(self) -> dict:
        return {"field_name": self.name, "index": self.indexed, "unique": self.unique}

    def _add_value(self, document_key, value):
        hash_value = hash(value)
        if self.unique and len(self._hash_index[hash_value]) > 0:
            raise UniqueViolation
        self._hash_index[hash_value].add(document_key)

    def _remove_value(self, document_key, value):
        hash_value = hash(value)
        self._hash_index[hash_value].discard(document_key)

    def _get_keys(self, comparison_type: ComparisonType, value):
        if comparison_type == ComparisonType.EQUAL:
            hash_value = hash(value)
            return self._hash_index[hash_value]

    def __eq__(self, other):
        return Query(field=self, comparison_type=ComparisonType.EQUAL, value=other)

    def __ne__(self, other):
        return Query(field=self, comparison_type=ComparisonType.NOT_EQUAL, value=other)

    def __repr__(self):
        return "Field(%r, index=%r, unique=%r)" % (self.name, self.indexed, self.unique)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['smolstore']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'smolstore',
    'version': '0.2.1a0',
    'description': 'A simple in-memory Python document store',
    'long_description': None,
    'author': 'Raymond Botha',
    'author_email': 'botha@riseup.net',
    'url': 'https://github.com/raybotha/smolstore',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
from typing import Iterable
from typing import Mapping
from typing import MutableMapping

from .field import Field


class Fields(Mapping):
    def __init__(self, _fields: Iterable[Field] = None):
        if not _fields:
            _fields = {}
        self.__dict__ = {field.name: field for field in _fields}

    @classmethod
    def _deserialize(cls, _data: MutableMapping):
        fields = cls()
        fields.__dict__ = {
            field_name: Field._deserialize(field) for field_name, field in _data.items()
        }
        return fields

    def _serialize(self) -> dict:
        return {field_name: field._serialize() for field_name, field in self.items()}

    def _register(self, document_key, field_name, old_value, new_value):
        field = self.get(field_name)
        if field is None:
            self[field_name] = Field(field_name)
        else:
            if field.indexed:
                field._add_value(document_key, new_value)
                field._remove_value(document_key, old_value)

    def _unregister(self, document_key, field_name, value):
        field = self.get(field_name)
        if field and field.indexed:
            field._remove_value(document_key, value)

    def __setitem__(self, key, value):
        self.__dict__.__setitem__(key, value)

    def __getitem__(self, key) -> Field:
        return self.__dict__.__getitem__(key)

    def __iter__(self):
        return self.__dict__.__iter__()

    def __len__(self):
        return self.__dict__.__len__()

    def __repr__(self):
        return "Fields(%r)" % list(self.__dict__.values())
